""" Module to access and control ublox gnss receivers """
# pylint: disable=C0103
import time
import datetime

from nmubxlib.server_tty import GnssUBlox # pylint: disable=E0401
from nmubxlib.ubx_upd_sos import UbxUpdSosPoll, UbxUpdSosAction # pylint: disable=E0401
from nmubxlib.ubx_cfg_rst import UbxCfgRstAction # pylint: disable=E0401
from nmubxlib.ubx_cfg_nav5 import UbxCfgNav5Poll # pylint: disable=E0401
from nmubxlib.ubx_cfg_navx5 import UbxCfgNavx5Poll # pylint: disable=E0401
from nmubxlib.ubx_cfg_nmea import UbxCfgNmeaPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_esfla import UbxCfgEsflaPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_esfalg import UbxCfgEsfAlgPoll # pylint: disable=E0401
from nmubxlib.ubx_mon_ver import UbxMonVerPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_prt import UbxCfgPrtPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_cfg import UbxCfgCfgAction # pylint: disable=E0401
from nmubxlib.ubx_cfg_rate import UbxCfgRatePoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_hnr import UbxCfgHnrPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_gnss import UbxCfgGnssPoll # pylint: disable=E0401
from nmubxlib.ubx_mga_ini_time_utc import UbxMgaIniTimeUtc # pylint: disable=E0401
from nmubxlib.ubx_cfg_tp5 import UbxCfgTp5Poll  # pylint: disable=E0401
from nmubxlib.cfgkeys import UbxKeyId, CfgKeyValues # pylint: disable=E0401
from nmubxlib.ubx_cfg_valget import UbxCfgValGetPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_valset import UbxCfgValSetAction # pylint: disable=E0401

from .logger import log
from .ubx_utils import UbxUtils, InvalidValue


MICROSECONDS_PER_SECOND = 1000000
NEO_M9_PROTVER = "35.10"

class UnsupportedConfiguration(Exception):
    """ The requested configuration is not supported by this device """

class InvalidNmeaVersion(Exception):
    """ Raised when trying to set an NMEA version which is not supported """

class ProtocolVersionNotFound(Exception):
    """ The protocol version could not be read. Not a ublox receiver ? """

class PollingException(Exception):
    """ Raised  when polling data from the receiver fails """

class UbxSetupFailed(Exception):
    """ Raised when initial setup failed"""

class ConfigurationException(Exception):
    """ Raised when applying a configuration to the receiver failed """

class UnexpectedResponseException(Exception):
    """ Raised when the receiver answer with an unexpected answer """

class UbloxReceiver():
    """ Class to access and control ublox gnss receivers """

    TP5_FLAG_ACTIVE = (1 << 0)
    TP5_FLAG_LOCK_GPS_FREQ = (1 << 1)
    TP5_FLAG_IS_LEN = (1 << 4)

    def __init__(self, tty, bitrate, configure_bitrate=True):
        self.tty = tty
        self.bitrate = bitrate
        self.ubx = None
        self.ubx_protocol_version = None


        while True:
            try:
                # Setup ubx accessor
                self.__setup_ubx(bitrate, configure_bitrate)
                break
            except UbxSetupFailed:
                if not configure_bitrate:
                    return
                time.sleep(0.5)


    def __get_current_bitrate(self):
        log.info("Detecting bitrate")
        # 115200 is the one we want to use
        # 38400 is the default of NEO-M9
        # 9600 us the default of NEO-M8
        BITRATES_TO_TEST = [ 115200, 38400, 9600 ]
        while True:
            for bitrate in BITRATES_TO_TEST:
                self.ubx.set_baudrate(bitrate)
                if self.ubx.scan():
                    log.info("Bitrate detected: %d", bitrate)
                    return bitrate

            # Bitrate detection failed, waiting and retrying
            time.sleep(0.5)


    def __read_and_setup_bitrate(self, bitrate):
        crt_bitrate = self.__get_current_bitrate()
        while crt_bitrate != bitrate:
            # Configure requested bitrate
            self.config_com_bitrate(bitrate)
            log.warning("Receiver bitrate was %d, changed to %d", crt_bitrate, bitrate)
            # Check that bitrate has really changed
            crt_bitrate = self.__get_current_bitrate()


    def __setup_ubx(self, bitrate, configure_bitrate):
        self.ubx = GnssUBlox(self.tty, bitrate)
        ready = self.ubx.setup()
        if not ready:
            self.ubx.cleanup()
            del self.ubx
            self.ubx = None
            raise UbxSetupFailed("UBX setup failed")

        # Setup bitrate
        if configure_bitrate:
            self.__read_and_setup_bitrate(bitrate)

        # Try to poll data from receiver
        err_cnt = 0
        while err_cnt < 3:
            try:
                self.ubx_protocol_version = self.get_ubx_protocol_version()
                log.info("Connected to GNSS receiver")
                return
            except PollingException:
                log.warning("Getting protocol version failed, retrying")
                err_cnt += 1
                time.sleep(0.5)

        self.ubx.cleanup()
        self.ubx = None
        raise UbxSetupFailed("Poll not working")



    def cleanup(self):
        if self.ubx is None:
            return
        self.ubx.cleanup()
        log.debug("Communication to GNSS receiver cleaned up")
        del self.ubx
        self.ubx = None


    # ---------------------------------------------------------------------------------------
    #   Helper Functions
    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------
    def __poll(self, poller):
        err_cnt = 0
        while err_cnt < 3:
            cfg = self.ubx.poll(poller)
            if cfg is None:
                err_cnt += 1
                time.sleep(0.5)
            else:
                return cfg

        if cfg is None:
            raise PollingException("Polling failed for " + str(poller))


    # ---------------------------------------------------------------------------------------

    def __poll_config_values(self, config_names):
        """
        Polls the request configs
        The returned list can be modified and set to the receiver using __set_config_values
        """

        poll_signals = UbxCfgValGetPoll(config_names)
        cfg = self.__poll(poll_signals)
        if cfg is None:
            log.error(f"Polling requested signals failed: {config_names}")
            raise ConfigurationException()

        cfg_list = []
        for counter,_ in enumerate(config_names):
            cfg_list.append(cfg.get(f'data{counter}'))

        return cfg_list

    def __set_config_values(self, cfg_list):
        """
        Sets the requested configs to the receiver
        the config list should usually be the list returned by __poll_config_values.

        The values can be modified like this:
        cfg[0].value = new_value
        """
        cfg_setval = UbxCfgValSetAction(cfg_list)

        writeResp = self.ubx.set(cfg_setval)
        if writeResp is None:
            log.error(f"nmubxlib response failure when configuring : {cfg_list}, {writeResp}")
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------

    def _poll_gnss_cfg(self):
        return self.__poll(UbxCfgGnssPoll())

    def _poll_update_rate(self):
        return self.__poll(UbxCfgRatePoll())

    def _poll_hnr(self):
        return self.__poll(UbxCfgHnrPoll())

    def _poll_port(self):
        port_poller = UbxCfgPrtPoll()
        port_poller.f.PortId = UbxCfgPrtPoll.PORTID_Uart
        return self.__poll(port_poller)


    # ---------------------------------------------------------------------------------------
    def _poll_save_on_shutdown(self):
        return self.__poll(UbxUpdSosPoll())


    # ---------------------------------------------------------------------------------------
    def _poll_nav_engine(self):
        return self.__poll(UbxCfgNav5Poll())


    # ---------------------------------------------------------------------------------------
    def _poll_navx5(self):
        return self.__poll(UbxCfgNavx5Poll())


    # ---------------------------------------------------------------------------------------
    def _poll_nav_protocol(self):
        return self.__poll(UbxCfgNmeaPoll())


    # ---------------------------------------------------------------------------------------
    def _poll_positioning(self):
        return self.__poll(UbxCfgEsflaPoll())


    # ---------------------------------------------------------------------------------------
    def _poll_alignment(self):
        return self.__poll(UbxCfgEsfAlgPoll())


    # ---------------------------------------------------------------------------------------
    def _poll_information(self):
        return self.__poll(UbxMonVerPoll())


    # ---------------------------------------------------------------------------------------
    def _sos_action(self, action):
        sos = UbxUpdSosAction()
        sos.f.cmd = action
        writeResp = self.ubx.set(sos)
        if writeResp is None:
            log.error("Not reponse from nmubxlib when saving or clearing receiver state: " + str(writeResp))
            raise ConfigurationException("Could not save or clear receiver state")

    def save_receiver_state(self):
        self._sos_action(UbxUpdSosAction.SAVE)
        log.info("Saving receiver state successfully performed")

    def clear_receiver_state(self):
        self._sos_action(UbxUpdSosAction.CLEAR)
        log.info("Clearing receiver state successfully performed")

    # ---------------------------------------------------------------------------------------
    def verify_restoring_state(self):
        sos_state = self._poll_save_on_shutdown()
        cmdField = sos_state.f.cmd
        response = sos_state.f.response
        if cmdField != 3:
            log.error("Restore verification got not the desired response: cmd=" + str(cmdField))
            raise UnexpectedResponseException("Unexpected response when verifying restore")
        state = "Unknown"
        if response == 1:
            state = "Failed"
            log.warning("Receiver restoring state: " + state)
        elif response == 2:
            state = "Success"
            log.info("Receiver restoring state: " + state)
        elif response == 3:
            state = "Nothing to restore"
            log.info("Receiver restoring state: " + state)
        else:
            log.warning("Receiver restoring state: " + state)


    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------
    #   Main Ublox Functions
    # ---------------------------------------------------------------------------------------
    def stop_gnss_function(self):
        rst = UbxCfgRstAction()
        rst.stop()
        rst.pack()
        self.ubx.fire_and_forget(rst)
        log.info("Stopping gnss function")

    def start_gnss_function(self):
        rst = UbxCfgRstAction()
        rst.start()
        rst.pack()
        self.ubx.fire_and_forget(rst)

    def coldboot_receiver(self):
        rst = UbxCfgRstAction()
        rst.cold_start()
        rst.pack()
        self.ubx.fire_and_forget(rst)
        log.info("Cold boot of GNSS receiver triggered, let receiver start")
        time.sleep(5)


    # ---------------------------------------------------------------------------------------
    def reset_config(self):
        cfgrst = UbxCfgCfgAction()
        cfgrst.reset(UbxCfgCfgAction.MASK_All)
        cfgrst.pack()
        self.ubx.fire_and_forget(cfgrst)
        log.info("Reset GNSS receiver configuration to default, let receiver start with default config")
        time.sleep(2)


    # ---------------------------------------------------------------------------------------
    def config_com_bitrate(self, bitrate):
        portcfg = self._poll_port()
        log.info("Setting communication bitrate from " + str(portcfg.f.baudRate) + " to " + str(bitrate))
        portcfg.f.baudRate = bitrate

        # Note: usually we would use ubx.set() to send the request and wait for an ack
        # Unfortunately the modem changes the bitrate immediately so the ACK is not
        # received.
        portcfg.pack()
        self.ubx.fire_and_forget(portcfg)
        log.info("Set default bitrate to " + str(bitrate) + " bit/s")

    # ---------------------------------------------------------------------------------------
    def set_receiver_init_time(self):
        mgainitime = UbxMgaIniTimeUtc()
        dt = datetime.datetime.utcnow()
        mgainitime.set_datetime(dt)
        self.ubx.fire_and_forget(mgainitime)
        log.info(f"Initial time (UTC): {dt} set to receiver")

    # ---------------------------------------------------------------------------------------
    def get_ubx_protocol_version(self):
        """ Returns protocol version as integer """
        rawinfo = self._poll_information()
        extension = 1
        while True:
            try:
                field = rawinfo.get(f'extension_{extension}').pack().decode('utf8').strip('\x00')
                if 'PROTVER=' in field:
                    return int(field.split('=')[1].replace('.',''))
                extension += 1

            except AttributeError as exc:
                # End of the extensions
                log.error(exc)
                raise ProtocolVersionNotFound() from exc


    def is_ubx_protocol_version_greater_or_equal_to(self, version):
        return self.ubx_protocol_version >= int(version.replace('.',''))

    def get_info(self):
        model = "n/a"
        fw = "n/a"
        ubxprot = "n/a"
        sat = "n/a"
        svc = "n/a"
        sw = "n/a"
        hw = "n/a"
        try:
            rawinfo = self._poll_information()
            model = rawinfo.f.extension_3.split("=")[1]
            fw = rawinfo.f.extension_1.split("=")[1]
            ubxprot = rawinfo.f.extension_2.split("=")[1]
            if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
                sat = rawinfo.f.extension_4
                svc = rawinfo.f.extension_5
            else:
                # Older systems had an additional FIS entry in 4th extension
                sat = rawinfo.f.extension_5
                svc = rawinfo.f.extension_6
            sw = rawinfo.f.swVersion
            hw = rawinfo.f.hwVersion
        except:
            log.error("Poll-Error: Cannot get information")

        validity = UbxUtils.get_fw_label(model, fw).value
        if validity:
            fw = f"{fw} ({validity})"

        info = {
            "vendor": "u-blox",
            "model": model,
            "fw": fw,
            "ubxprot": ubxprot,
            "sat": sat,
            "svc": svc,
            "sw": sw,
            "hw": hw
        }
        return info

    # ---------------------------------------------------------------------------------------
    def set_alignment(self, yaw, pitch, roll):
        """ input in degrees, parsing, converting to centridegrees and calling the right setters """
        if yaw is not None:
            try:
                UbxUtils.range_check(yaw, 0, 360)
                yaw = yaw * 100
            except InvalidValue:
                log.error("yaw is not valid, not setting it: %d", yaw)
                yaw = None
        if pitch is not None:
            try:
                UbxUtils.range_check(pitch, -90, 90)
                pitch = pitch * 100
            except InvalidValue:
                log.error("pitch is not valid, not setting it: %d", pitch)
                pitch = None
        if roll is not None:
            try:
                UbxUtils.range_check(roll, -180, 180)
                roll = roll * 100
            except InvalidValue:
                log.error("roll is not valid, not setting it: %d", roll)
                roll = None

        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            self._set_alignment_v2(yaw, pitch, roll)
        else:
            self._set_alignment_v1(yaw, pitch, roll)

    def _set_alignment_v2(self, yaw, pitch, roll):
        """ ubx protocol >= 35.10, angles in centidegrees """
        configs = self.__poll_config_values([
            UbxKeyId.CFG_SFIMU_AUTO_MNTALG_ENA,
            UbxKeyId.CFG_SFIMU_IMU_MNTALG_YAW,
            UbxKeyId.CFG_SFIMU_IMU_MNTALG_PITCH,
            UbxKeyId.CFG_SFIMU_IMU_MNTALG_ROLL
        ])

        if yaw is not None or pitch is not None or roll is not None:
            # Make sure auto-alignment is off if some angles are manually configured
            configs[0].value = False

        if yaw is not None:
            configs[1].value = yaw
        if pitch is not None:
            configs[2].value = pitch
        if roll is not None:
            configs[3].value = roll

        self.__set_config_values(configs)

    def _set_alignment_v1(self, yaw, pitch, roll):
        """ ubx protocol < 35.10, angles in centidegrees """
        aligncfg = self._poll_alignment()

        if yaw is not None:
            aligncfg.f.yaw = yaw
        if pitch is not None:
            aligncfg.f.pitch = pitch
        if roll is not None:
            aligncfg.f.roll = roll

        writeResp = self.ubx.set(aligncfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring alignment: " + str(writeResp))
            raise ConfigurationException()

    def get_alignment(self):
        """ yaw, pitch, roll in degrees """
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            yaw, pitch, roll = self._get_alignment_v2()
        else:
            yaw, pitch, roll = self._get_alignment_v1()
        return yaw / 100, pitch / 100, roll / 100

    def _get_alignment_v2(self):
        """ ubx protocol >= 35.10, angles in centidegrees """
        configs = self.__poll_config_values([UbxKeyId.CFG_SFIMU_IMU_MNTALG_YAW, UbxKeyId.CFG_SFIMU_IMU_MNTALG_PITCH, UbxKeyId.CFG_SFIMU_IMU_MNTALG_ROLL])
        return configs[0].value, configs[1].value, configs[2].value

    def _get_alignment_v1(self):
        """ ubx protocol < 35.10, angles in centidegrees """
        cfg = self._poll_alignment()

        return cfg.f.yaw, cfg.f.pitch, cfg.f.roll


    # ---------------------------------------------------------------------------------------
    def set_positioning(self, vrp2antenna, vrp2imu):
        """ Inputs: arrays of 3 (x,y,z) in meters  """
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            self._set_positioning_v2(vrp2antenna, vrp2imu)
        else:
            self._set_positioning_v1(vrp2antenna, vrp2imu)

    def _set_positioning_v2(self, vrp2antenna, vrp2imu):
        """ ubx protocol >= 35.10, values in m converted to cm for receiver """
        configs = self.__poll_config_values([
            UbxKeyId.CFG_SFODO_IMU2VRP_LA_X,
            UbxKeyId.CFG_SFODO_IMU2VRP_LA_Y,
            UbxKeyId.CFG_SFODO_IMU2VRP_LA_Z,
            UbxKeyId.CFG_SFIMU_IMU2ANT_LA_X,
            UbxKeyId.CFG_SFIMU_IMU2ANT_LA_Y,
            UbxKeyId.CFG_SFIMU_IMU2ANT_LA_Z,
            ])

        imu2vrp = UbxUtils.get_imu2vrp(vrp2imu)
        imu2ant = UbxUtils.get_imu2antenna(vrp2imu=vrp2imu, vrp2antenna=vrp2antenna)

        configs[0].value = round(imu2vrp[0] * 100)
        configs[1].value = round(imu2vrp[1] * 100)
        configs[2].value = round(imu2vrp[2] * 100)

        configs[3].value = round(imu2ant[0] * 100)
        configs[4].value = round(imu2ant[1] * 100)
        configs[5].value = round(imu2ant[2] * 100)

        self.__set_config_values(configs)

    def _set_positioning_v1(self, vrp2antenna, vrp2imu):
        """ Values in m converted to cm for receiver """
        poscfg = self._poll_positioning()

        if vrp2antenna is not None:
            index = UbxUtils.find_arm_by_name(poscfg, 'vrp2antenna')
            poscfg.f.__setattr__(f'leverArmX_{index}', round(vrp2antenna[0] * 100))
            poscfg.f.__setattr__(f'leverArmY_{index}', round(vrp2antenna[1] * 100))
            poscfg.f.__setattr__(f'leverArmZ_{index}', round(vrp2antenna[2] * 100))

        if vrp2imu is not None:
            index = UbxUtils.find_arm_by_name(poscfg, 'vrp2imu')
            poscfg.f.__setattr__(f'leverArmX_{index}', round(vrp2imu[0] * 100))
            poscfg.f.__setattr__(f'leverArmY_{index}', round(vrp2imu[1] * 100))
            poscfg.f.__setattr__(f'leverArmZ_{index}', round(vrp2imu[2] * 100))

        writeResp = self.ubx.set(poscfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring positioning: " + str(writeResp))
            raise ConfigurationException()

    def get_positioning(self):
        """ Returns positions as a tuple of arrays (both x,y,z in meters): vrp2antenna, vrp2imu """
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            return self._get_positioning_v2()
        else:
            return self._get_positioning_v1()

    def _get_positioning_v2(self):
        """ Returns positions as a tuple of arrays (both x,y,z in meters): vrp2antenna, vrp2imu """
        configs = self.__poll_config_values([
            UbxKeyId.CFG_SFODO_IMU2VRP_LA_X,
            UbxKeyId.CFG_SFODO_IMU2VRP_LA_Y,
            UbxKeyId.CFG_SFODO_IMU2VRP_LA_Z,
            UbxKeyId.CFG_SFIMU_IMU2ANT_LA_X,
            UbxKeyId.CFG_SFIMU_IMU2ANT_LA_Y,
            UbxKeyId.CFG_SFIMU_IMU2ANT_LA_Z,
            ])

        imu2vrp = [
                configs[0].value / 100.,
                configs[1].value / 100.,
                configs[2].value / 100.
                ]

        imu2antenna = [
                configs[3].value / 100.,
                configs[4].value / 100.,
                configs[5].value / 100.
                ]

        vrp2antenna = UbxUtils.get_vrp2antenna(imu2vrp=imu2vrp, imu2antenna=imu2antenna)
        vrp2imu = UbxUtils.get_vrp2imu(imu2vrp)

        return vrp2antenna, vrp2imu

    def _get_positioning_v1(self):
        """ Returns positions as a tuple of arrays (both x,y,z in meters): vrp2antenna, vrp2imu """
        vrp2antenna = []
        vrp2imu = []

        poscfg = self._poll_positioning()

        index = UbxUtils.find_arm_by_name(poscfg, 'vrp2antenna')
        vrp2antenna.append(poscfg.f.get(f'leverArmX_{index}').value / 100.)
        vrp2antenna.append(poscfg.f.get(f'leverArmY_{index}').value / 100.)
        vrp2antenna.append(poscfg.f.get(f'leverArmZ_{index}').value / 100.)

        index = UbxUtils.find_arm_by_name(poscfg, 'vrp2imu')
        vrp2imu.append(poscfg.f.get(f'leverArmX_{index}').value / 100.)
        vrp2imu.append(poscfg.f.get(f'leverArmY_{index}').value / 100.)
        vrp2imu.append(poscfg.f.get(f'leverArmZ_{index}').value / 100.)

        return vrp2antenna, vrp2imu


    # ---------------------------------------------------------------------------------------

    def get_nmea_version(self):
        """ Returns nmea version as floating point """
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            return self._get_nmea_version_v2()
        else:
            return self._get_nmea_version_v1()

    @staticmethod
    def _get_nmea_version2value_map():
        """ Maps floating point version number to ubx constants """
        return {
                2.1: 21,
                2.3: 23,
                4.0: 40,
                4.10: 41,
                4.11: 42
        }
    def _get_nmea_version_v2(self):
        version2value_map = self._get_nmea_version2value_map()
        value2version_map = {v: k for k, v in version2value_map.items()}

        configs = self.__poll_config_values([UbxKeyId.CFG_NMEA_PROTVER])
        return value2version_map[configs[0].value]

    def _get_nmea_version_v1(self):
        protcfg = self._poll_nav_protocol()
        vProtRead = int(protcfg.f.nmeaVersion)
        vProtReadMain = round(float(vProtRead) / 16)
        vProtReadSub = float(vProtRead - (vProtReadMain * 16))
        vProtReadSub = vProtReadSub / 10
        return float(float(vProtReadMain) + vProtReadSub)

    def set_nmea_version(self, version):
        """ Version is a floating point number """
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            self._set_nmea_version_v2(version)
        else:
            self._set_nmea_version_v1(version)

    def _set_nmea_version_v2(self, version):
        """ Version is a floating point number """
        version2value = self._get_nmea_version2value_map()
        try:
            value = version2value[version]
        except KeyError as exc:
            log.warning("Unsupported NMEA version")
            raise InvalidNmeaVersion() from exc
        configs = self.__poll_config_values([UbxKeyId.CFG_NMEA_PROTVER])
        configs[0].value = value
        self.__set_config_values(configs)

    def _set_nmea_version_v1(self, version):
        """ Version is a floating point number """
        protcfg = self._poll_nav_protocol()
        vProt = version
        vProtInt = round(vProt)
        vProtMain = vProtInt * 16
        vProtSub = round((vProt - vProtInt) * 10)
        protcfg.f.nmeaVersion = round(vProtMain + vProtSub)

        writeResp = self.ubx.set(protcfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring navigation protocol: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------
    def set_nav_engine(self, dyn_model=None, fix_mode=None):
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            self._set_nav_engine_v2(dyn_model, fix_mode)
        else:
            self._set_nav_engine_v1(dyn_model, fix_mode)

    def _set_nav_engine_v2(self, dyn_model=None, fix_mode=None):
        configs = self.__poll_config_values([UbxKeyId.CFG_NAVSPG_DYNMODEL, UbxKeyId.CFG_NAVSPG_FIXMODE])
        if dyn_model is not None:
            try:
                configs[0].value = UbxUtils.nav_engine_model_string2value(dyn_model)
            except InvalidValue:
                log.error("mode value is wrong, not setting it: %s", dyn_model)
        if fix_mode is not None:
            try:
                configs[1].value = UbxUtils.get_nav_engine_fix_mode_value(fix_mode)
            except InvalidValue:
                log.error("fix_mode value is wrong, not setting it: %s", fix_mode)
        self.__set_config_values(configs)



    def _set_nav_engine_v1(self, dyn_model=None, fix_mode=None):
        navcfg = self._poll_nav_engine()
        if dyn_model is not None:
            try:
                navcfg.f.dynModel = UbxUtils.nav_engine_model_string2value(dyn_model)
            except InvalidValue:
                log.error("mode value is wrong, not setting it: %s", dyn_model)
        if fix_mode is not None:
            try:
                navcfg.f.fixMode = UbxUtils.get_nav_engine_fix_mode_value(fix_mode)
            except InvalidValue:
                log.error("fix_mode value is wrong, not setting it: %s", fix_mode)

        writeResp = self.ubx.set(navcfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring navigation engine: " + str(writeResp))
            raise ConfigurationException()

    def get_nav_engine(self):
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            return self._get_nav_engine_v2()
        else:
            return self._get_nav_engine_v1()

    def _get_nav_engine_v2(self):
        configs = self.__poll_config_values([UbxKeyId.CFG_NAVSPG_DYNMODEL, UbxKeyId.CFG_NAVSPG_FIXMODE])
        dyn_model = UbxUtils.nav_engine_model_value2string(configs[0].value)
        fix_mode = UbxUtils.get_nav_engine_fix_mode_string(configs[1].value)
        return dyn_model, fix_mode

    def _get_nav_engine_v1(self):
        navcfg = self._poll_nav_engine()
        dyn_model = UbxUtils.nav_engine_model_value2string(navcfg.f.dynModel)
        fix_mode = UbxUtils.get_nav_engine_fix_mode_string(navcfg.f.fixMode)
        return dyn_model, fix_mode

    # ---------------------------------------------------------------------------------------
    def set_navx5(self, adr_enable=None):
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            self._set_navx5_v2(adr_enable)
        else:
            self._set_navx5_v1(adr_enable)

    def _set_navx5_v2(self, adr_enable=None):
        configs = self.__poll_config_values([UbxKeyId.CFG_SFCORE_USE_SF])
        if adr_enable is not None:
            configs[0].value = adr_enable
            self.__set_config_values(configs)

    def _set_navx5_v1(self, adr_enable=None):
        navx5 = self._poll_navx5()
        navx5.f.mask2 = 0
        if adr_enable is not None:
            navx5.f.mask2 |= (1 << 6)
            navx5.f.useAdr = 1 if adr_enable else 0

        writeResp = self.ubx.set(navx5)
        if writeResp is None:
            log.error("nmubxlib response failure when enabling or disabling ADR: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------

    def get_update_rate(self):
        """ Returns update rate in Hz """
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            return self._get_update_rate_v2()
        else:
            return self._get_update_rate_v1()

    def _get_update_rate_v2(self):
        """ Returns update rate in Hz """
        configs = self.__poll_config_values([UbxKeyId.CFG_RATE_MEAS, UbxKeyId.CFG_RATE_NAV])
        return round(1000 / configs[0].value)  # ms to Hz

    def _get_update_rate_v1(self):
        """ Returns update rate in Hz """
        ratecfg = self._poll_update_rate()
        return round(1000 / ratecfg.f.measRate)  # ms to Hz

    def set_update_rate(self, rate):
        """ Update Rate in HZ """
        try:
            UbxUtils.range_check(rate, 1, 2)
        except InvalidValue:
            log.error("update rate value is wrong, not setting it: %d", rate)
            return

        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            self._set_update_rate_v2(rate)
        else:
            self._set_update_rate_v1(rate)

    def _set_update_rate_v2(self, rate):
        """ Update Rate in HZ, version for protocol >= 35.10 """
        configs = self.__poll_config_values([UbxKeyId.CFG_RATE_MEAS, UbxKeyId.CFG_RATE_NAV])
        configs[0].value = round(1000 / rate)  # Hz to ms
        configs[1].value = 1 # cycle. ie synced with measRate. 2 cycle means, every 2 measurement
        self.__set_config_values(configs)

    def _set_update_rate_v1(self, rate):
        """ Update Rate in HZ, version for protocol < 35.10 """
        ratecfg = self._poll_update_rate()
        ratecfg.f.measRate = round(1000 / rate)  # Hz to ms
        ratecfg.f.navRate = 1 # cycle. ie synced with measRate. 2 cycle means, every 2 measurement
        writeResp = self.ubx.set(ratecfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring update rate: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------

    def get_high_navigation_rate(self):
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            return self._get_update_rate_v2()
        else:
            return self._get_update_rate_v1()

    def _get_high_navigation_rate_v2(self):
        configs = self.__poll_config_values([UbxKeyId.CFG_RATE_NAV_PRIO])
        return configs[0].value

    def _get_high_navigation_rate_v1(self):
        cfg = self._poll_hnr()
        return cfg.get_rate_in_hz()

    def set_high_navigation_rate(self, rate):
        try:
            UbxUtils.range_check(rate, 1, 30)
        except InvalidValue:
            log.error("high navigation range out of range: %d", rate)
            return

        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            self._set_high_navigation_rate_v2(rate)
        else:
            self._set_high_navigation_rate_v1(rate)

    def _set_high_navigation_rate_v2(self, rate):
        configs = self.__poll_config_values([UbxKeyId.CFG_RATE_NAV_PRIO])
        configs[0].value = rate
        self.__set_config_values(configs)

    def _set_high_navigation_rate_v1(self, rate):
        """ High navigation rate """
        cfg = self._poll_hnr()
        cfg.set_rate_in_hz(rate)
        writeResp = self.ubx.set(cfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring update rate: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------
    def __switch_gnss_system(self, cfg, value, system):
        if value:
            cfg.enable_gnss(system)
        else:
            cfg.disable_gnss(system)

    def set_gnss_systems(self, gps=None, glonass=None, galileo=None, beidou=None, sbas=None, imes=None, qzss=None):
        """ Enable or disable sattelites positioning systems """
        if self.is_ubx_protocol_version_greater_or_equal_to("23.01"):
            self._set_gnss_systems_v2(gps, glonass, galileo, beidou, sbas, imes, qzss)
        else:
            self._set_gnss_systems_v1(gps, glonass, galileo, beidou, sbas, imes, qzss)

    def _set_gnss_systems_v2(self, gps=None, glonass=None, galileo=None, beidou=None, sbas=None, imes=None, qzss=None):
        """ Implementation for UBX protocol version >= 23.01 """

        # Read currently active GNSS systems
        configs = self.__poll_config_values([
            UbxKeyId.CFG_SIGNAL_GPS_ENA,
            UbxKeyId.CFG_SIGNAL_GPS_L1CA_ENA,
            UbxKeyId.CFG_SIGNAL_GLO_ENA,
            UbxKeyId.CFG_SIGNAL_GLO_L1_ENA,
            UbxKeyId.CFG_SIGNAL_GAL_ENA,
            UbxKeyId.CFG_SIGNAL_GAL_E1_ENA,
            UbxKeyId.CFG_SIGNAL_BDS_ENA,
            UbxKeyId.CFG_SIGNAL_BDS_B1_ENA,
            UbxKeyId.CFG_SIGNAL_SBAS_ENA,
            UbxKeyId.CFG_SIGNAL_SBAS_L1CA_ENA,
            UbxKeyId.CFG_SIGNAL_QZSS_ENA,
            UbxKeyId.CFG_SIGNAL_QZSS_L1CA_ENA,
            UbxKeyId.CFG_SIGNAL_QZSS_L1S_ENA
        ])

        if gps is not None:
            configs[0].value = gps
            configs[1].value = gps

        if glonass is not None:
            configs[2].value = glonass
            configs[3].value = glonass

        if galileo is not None:
            configs[4].value = galileo
            configs[5].value = galileo

        if beidou is not None:
            configs[6].value = beidou
            configs[7].value = beidou

        if sbas is not None:
            configs[8].value = sbas
            configs[9].value = sbas

        if qzss is not None:
            configs[10].value = qzss
            configs[11].value = qzss
            configs[12].value = qzss

        if imes is not None:
            # NB: This function is called whenever a recent version of the ubx protocol is used
            # But we cannot rely on the protocol version to know which gnss systems are available
            # It just works for the moment because we have neo-m8 with IMES and old protocol
            # And neo-m9 with new protocol and without IMES
            log.info("IMES cannot be configured on this device, ignoring")
            if imes:
                log.error("IMES is not available on this device")
                raise UnsupportedConfiguration("IMES is not available on this device")

        self.__set_config_values(configs)

    def _set_gnss_systems_v1(self, gps=None, glonass=None, galileo=None, beidou=None, sbas=None, imes=None, qzss=None):
        """ Implementation for UBX protocol version < 23.01 """
        cfg = self._poll_gnss_cfg()
        if gps is not None:
            self.__switch_gnss_system(cfg, gps, cfg.GNSS_GPS)
        if galileo is not None:
            self.__switch_gnss_system(cfg, galileo, cfg.GNSS_Galileo)
        if beidou is not None:
            self.__switch_gnss_system(cfg, beidou, cfg.GNSS_BeiDou)
        if glonass is not None:
            self.__switch_gnss_system(cfg, glonass, cfg.GNSS_GLONASS)
        if sbas is not None:
            self.__switch_gnss_system(cfg, sbas, cfg.GNSS_SBAS)
        if imes is not None:
            self.__switch_gnss_system(cfg, imes, cfg.GNSS_IMES)
        if qzss is not None:
            self.__switch_gnss_system(cfg, qzss, cfg.GNSS_QZSS)

        writeResp = self.ubx.set(cfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring gnss systems: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------

    def __poll_time_pulse(self):
        poller = UbxCfgTp5Poll()
        # Although NEO-M8L has only one timepulse, we must refer to it as
        # TIMEPULSE2 (id = 1), according to datasheet
        poller.f.tpIdx = 1

        cfg = self.ubx.poll(poller)
        if cfg is None:
            log.error("Polling timepulse config failed")
            raise ConfigurationException()

        return cfg


    def set_timepulse(self, enable, frequency=None):
        """ frequency [Hz] """
        if not enable and frequency is None:
            frequency = 1

        period_us = round(MICROSECONDS_PER_SECOND / frequency)

        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            self._set_timepulse_v2(enable, period_us)
        else:
            self._set_timepulse_v1(enable, period_us)

    def _set_timepulse_v2(self, enable, period_us):
        cfg_keyvals = CfgKeyValues.from_keyvalues([
            (UbxKeyId.CFG_TP_TP2_ENA, enable),                  # Enable / disable
            (UbxKeyId.CFG_TP_PULSE_DEF, 0),                     # Specify pulse interval in us
            (UbxKeyId.CFG_TP_PULSE_LENGTH_DEF, 1),              # Specify pulse lengths in us
            (UbxKeyId.CFG_TP_POL_TP2, 1),                       # Rising edge at top of second

            (UbxKeyId.CFG_TP_PERIOD_TP2, period_us),            # full period
            (UbxKeyId.CFG_TP_LEN_TP2, round(period_us / 2)),      # 50% duty cycle

            (UbxKeyId.CFG_TP_USE_LOCKED_TP2, True),             # Use following settings when locked
            (UbxKeyId.CFG_TP_ALIGN_TO_TOW_TP2, True),           # Align pulse to top of seconds
            (UbxKeyId.CFG_TP_TIMEGRID_TP2, 1),                  # Lock to GPS time
            (UbxKeyId.CFG_TP_PERIOD_LOCK_TP2, period_us),       # full period when locked
            (UbxKeyId.CFG_TP_LEN_LOCK_TP2, round(period_us / 2)), # 50% duty cycle when locked
        ])

        self.__set_config_values(cfg_keyvals)

    def _set_timepulse_v1(self, enable, period_us):
        cfg = self.__poll_time_pulse()

        if enable is not None:
            if enable:
                cfg.f.flags |= self.TP5_FLAG_ACTIVE
                cfg.f.flags |= self.TP5_FLAG_LOCK_GPS_FREQ
            else:
                cfg.f.flags &= ~self.TP5_FLAG_ACTIVE
                cfg.f.flags &= ~self.TP5_FLAG_LOCK_GPS_FREQ

        if period_us is not None:
            cfg.f.flags |= self.TP5_FLAG_IS_LEN
            cfg.f.freqPeriod = period_us
            cfg.f.pulseLenRatio = round(period_us / 2)

        writeResp = self.ubx.set(cfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring time pulse")
            raise ConfigurationException()

    def get_timepulse(self):
        if self.is_ubx_protocol_version_greater_or_equal_to(NEO_M9_PROTVER):
            return self._get_timepulse_v2()
        else:
            return self._get_timepulse_v1()

    def _get_timepulse_v2(self):
        configs = self.__poll_config_values([UbxKeyId.CFG_TP_TP2_ENA, UbxKeyId.CFG_TP_PERIOD_TP2, UbxKeyId.CFG_TP_PULSE_DEF])
        enable = configs[0].value
        if configs[2].value == 1:
            # Reading frequency directly
            frequency = configs[1].value
        else:
            # Reading period
            frequency = round(MICROSECONDS_PER_SECOND / configs[1].value)

        return enable, frequency

    def _get_timepulse_v1(self):
        cfg = self.__poll_time_pulse()
        enable = True if cfg.f.flags & self.TP5_FLAG_ACTIVE != 0 else False
        if cfg.f.flags & self.TP5_FLAG_IS_LEN != 0:
            freq = round(MICROSECONDS_PER_SECOND / cfg.f.freqPeriod)
        else:
            freq = cfg.f.freqPeriod
        return enable, freq


    # ---------------------------------------------------------------------------------------

    def persist_receiver_config(self):
        ubx_cfg_cfg = UbxCfgCfgAction()
        ubx_cfg_cfg.f.saveMask = ubx_cfg_cfg.MASK_All
        log.info("Persisting receiver configuration")
        self.ubx.set(ubx_cfg_cfg)
