
import re
from enum import Enum

from nmubxlib.ubx_cfg_esfalg import UbxCfgEsfAlg  # pylint: disable=E0401
from nmubxlib.ubx_cfg_esfgawt import UbxCfgEsfGawt  # pylint: disable=E0401

from .logger import log


class ArmLeverNotFound(Exception):
    """ Raised if a configured arm lever is not found in the config of the receiver """

class InvalidValue(Exception):
    """ Raised when config file does not contain a valid value """

class UbxUtils():

    """
    Everything older than min_version will be marked as "obsolete"
    Everything older than min_recommended will be marked as "deprecated"
    Everything new than max_version will be marked as "untested"
    """
    SUPPORTED_FIRMWARE_VERSIONS = {
        "NEO-M9V": {
            "type": "MDR",
            "min_version": "2.10",
            "min_recommended": "2.10",
            "max_version": "2.10"
        },
        "NEO-M8L-0": {
            "type": "ADR",
            "min_version": "4.21",
            "min_recommended": "4.31",
            "max_version": "4.31",
        }
    }

    class FwLabels(Enum):
        UNTESTED = "Untested"
        DEPRECATED = "Deprecated"
        OBSOLETE = "Obsolete"
        VALID = None


    @staticmethod
    def get_fw_label(model, firmware_str):
        """ The label indicates if the firmware is valid, deprecated, obsolete or untested """
        try:
            supported_fw_versions = UbxUtils.SUPPORTED_FIRMWARE_VERSIONS[model]
        except KeyError:
            log.error(f"Unsupported u-blox receiver: {model}")
            return UbxUtils.FwLabels.UNTESTED


        regex_pattern_fw_info = re.compile('(?P<fw_type>\w*) (?P<fw_version>[\d.]*)')
        fw_info = regex_pattern_fw_info.search(firmware_str)
        if not fw_info:
            log.warning("unable to check if fw version is up to date")
            return UbxUtils.FwLabels.UNTESTED

        fw_type = fw_info.group("fw_type")
        fw_version = float(fw_info.group("fw_version"))

        if fw_type != supported_fw_versions['type']:
            log.warning("Unknown type of firmware for this model")
            return UbxUtils.FwLabels.UNTESTED

        if fw_version < float(supported_fw_versions['min_version']):
            return UbxUtils.FwLabels.OBSOLETE
        if fw_version < float(supported_fw_versions['min_recommended']):
            return UbxUtils.FwLabels.DEPRECATED
        if fw_version > float(supported_fw_versions['max_version']):
            return UbxUtils.FwLabels.UNTESTED

        return UbxUtils.FwLabels.VALID

    @staticmethod
    def get_vrp2imu(imu2vrp):
        """ Converts imu2vrp to vrp2imu"""
        return [ -imu2vrp[0], -imu2vrp[1], -imu2vrp[2] ]

    @staticmethod
    def get_imu2vrp(vrp2imu):
        """ Converts vrp2imu to imu2vrp """
        return [ -vrp2imu[0], -vrp2imu[1], -vrp2imu[2] ]

    @staticmethod
    def get_vrp2antenna(imu2vrp, imu2antenna):
        """ Converts imu2vrp and imu2antenna to vrp2antenna"""
        vrp2imu = UbxUtils.get_vrp2imu(imu2vrp=imu2vrp)
        x = vrp2imu[0] + imu2antenna[0]
        y = vrp2imu[1] + imu2antenna[1]
        z = vrp2imu[2] + imu2antenna[2]
        return [x, y ,z]

    @staticmethod
    def get_imu2antenna(vrp2imu, vrp2antenna):
        """ Converts vrp2imu and vrp2antenna to imu2antenna"""
        imu2vrp = UbxUtils.get_imu2vrp(vrp2imu=vrp2imu)
        x = imu2vrp[0] + vrp2antenna[0]
        y = imu2vrp[1] + vrp2antenna[1]
        z = imu2vrp[2] + vrp2antenna[2]
        return [x, y ,z]

    @staticmethod
    def find_arm_by_name(full_cfg, arm_name):
        arm_id = None

        # Convert arm name to arm number
        # Order is important: each arm has a matching id: 0 for vrp2antenna...
        arm_names = ['vrp2antenna', 'vrp2imu', 'imu2antenna', 'imu2vrp', 'imu2crp']
        for i, name in enumerate(arm_names):
            if arm_name == name:
                arm_id = i
                break

        if arm_id is None:
            raise ArmLeverNotFound()

        # Find position of arm with this id
        pos = 0
        while True:
            try:
                arm_id_cfg = full_cfg.f.get(f'leverArmType_{pos}').value
                if arm_id_cfg == arm_id:
                    return pos
                pos += 1
            except:
                raise ArmLeverNotFound()


    @staticmethod
    def range_check(value, min, max):
        if value > max or value < min:
            raise InvalidValue()


    @staticmethod
    def get_wheeltick_flags_string(flags):
        wtFlags = dict()
        wtFlags["useWtPin"] = 0
        if (flags & UbxCfgEsfGawt.MASK_useWtPin) != 0:
            wtFlags["useWtPin"] = 1
        # wtFlags["useWtSpeed"] = 0
        # if (flags & UbxCfgEsfGawt.MASK_useWtSpeed) != 0:
        #    wtFlags["useWtSpeed"] = 1
        return wtFlags


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def adapt_wheeltick_flags(actualFlags, flags):
        for flagItem in flags:
            if flagItem == "useWtPin":
                if flags[flagItem] == 0:
                    actualFlags &= ~UbxCfgEsfGawt.MASK_useWtPin
                else:
                    actualFlags |= UbxCfgEsfGawt.MASK_useWtPin
            elif flagItem == "useWtSpeed":
                if flags[flagItem] == 0:
                    actualFlags &= ~UbxCfgEsfGawt.MASK_useWtSpeed
                else:
                    actualFlags |= UbxCfgEsfGawt.MASK_useWtSpeed
            else:
                log.warning("Unsupported wheel tick flag item: " + str(flagItem))
        return actualFlags


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def get_alignment_bitfield_object(bitfield):
        bitfieldCfg = dict()
        bitfieldCfg["doAutoMntAlg"] = 0
        if (bitfield & UbxCfgEsfAlg.BITFIELD_doAutoMntAlg) != 0:
            bitfieldCfg["doAutoMntAlg"] = 1
        return bitfieldCfg


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def adapt_alignment_bitfield(actualField, field):
        for fieldItem in field:
            if fieldItem == "doAutoMntAlg":
                if field[fieldItem] == 0:
                    actualField &= ~UbxCfgEsfAlg.BITFIELD_doAutoMntAlg
                else:
                    actualField |= UbxCfgEsfAlg.BITFIELD_doAutoMntAlg
            else:
                log.warning("Unsupported alignment bitfield item: " + str(fieldItem))
        return actualField


    # ---------------------------------------------------------------------------------------

    DYNAMIC_MODELS = {
        0: "portable",
        2: "stationary",
        3: "pedestrian",
        4: "vehicle",
        5: "sea",
        6: "airborne <1g",
        7: "airborne <2g",
        8: "airborne <4g",
        9: "wrist worn watch",
        10: "bike",
        11: "mower",
        12: "e-scooter"
    }

    @staticmethod
    def nav_engine_model_value2string(value):
        """ Convert dynamic model from receiver's constant to human readable string """
        try:
            return UbxUtils.DYNAMIC_MODELS[value]
        except KeyError as exc:
            raise InvalidValue(f"Unsupported nav engine model value: {value}") from exc


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def nav_engine_model_string2value(model_string):
        """ Convert dynamic model from human readable string to receiver's constant """
        for key, item in UbxUtils.DYNAMIC_MODELS.items():
            if item.lower() == model_string.lower():
                return key

        raise InvalidValue(f"Unsupported nav engine model: {model_string}")


    # ---------------------------------------------------------------------------------------

    FIX_MODES = {
        1: "2D",
        2: "3D",
        3: "auto"
    }

    @staticmethod
    def get_nav_engine_fix_mode_string(value):
        """ Convert fix_mode from receiver's constant to human readable string """
        try:
            return UbxUtils.FIX_MODES[value]
        except KeyError as exc:
            raise InvalidValue(f"Unsupported nav engine fix mode: {value}") from exc


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def get_nav_engine_fix_mode_value(mode_string):
        """ Convert fix_mode from human readable string to receiver's constant """
        for key, value in UbxUtils.FIX_MODES.items():
            if value.lower() == mode_string.lower():
                return key

        raise InvalidValue(f"Unsupported nav engine fix mode: {mode_string}")
